/*
 * Intelbras PXE PCI Card Driver
 *
 * Based on previous works, designs, and archetectures conceived and
 * written by Jim Dixon <jim@lambdatel.com> and Mark Spencer <markster@digium.com>.
 *
 * Copyright (C) 2001 Jim Dixon / Zapata Telephony.
 * Copyright (C) 2001-2005, Digium, Inc.
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/time.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include "zaptel.h"
#ifdef LINUX26
#include <linux/moduleparam.h>
#endif
#include "pxe.h"

/*
 * Tasklets provide better system interactive response at the cost of the
 * possibility of losing a frame of data at very infrequent intervals.  If
 * you are more concerned with the performance of your machine, enable the
 * tasklets.  If you are strict about absolutely no drops, then do not enable
 * tasklets.
 */

/* #define ENABLE_TASKLETS */


/* Work queues are a way to better distribute load on SMP systems */
#ifdef LINUX26
/*
 * Work queues can significantly improve performance and scalability
 * on multi-processor machines, but requires bypassing some kernel
 * API's, so it's not guaranteed to be compatible with all kernels.
 */
/* #define ENABLE_WORKQUEUES */
#endif

/* Enable prefetching may help performance */
#define ENABLE_PREFETCH

/* Define to get more attention-grabbing but slightly more I/O using
   alarm status */
#define FANCY_ALARM

#define DEBUG_MAIN 		(1 << 0)
#define DEBUG_DTMF 		(1 << 1)
#define DEBUG_REGS 		(1 << 2)
#define DEBUG_TSI  		(1 << 3)
#define DEBUG_ECHOCAN 	(1 << 4)
#define DEBUG_RBS 		(1 << 5)
#define DEBUG_FRAMER		(1 << 6)

#ifdef ENABLE_WORKQUEUES
#include <linux/cpumask.h>

/* XXX UGLY!!!! XXX  We have to access the direct structures of the workqueue which
  are only defined within workqueue.c because they don't give us a routine to allow us
  to nail a work to a particular thread of the CPU.  Nailing to threads gives us substantially
  higher scalability in multi-CPU environments though! */

/*
 * The per-CPU workqueue (if single thread, we always use cpu 0's).
 *
 * The sequence counters are for flush_scheduled_work().  It wants to wait
 * until until all currently-scheduled works are completed, but it doesn't
 * want to be livelocked by new, incoming ones.  So it waits until
 * remove_sequence is >= the insert_sequence which pertained when
 * flush_scheduled_work() was called.
 */
 
struct cpu_workqueue_struct {

	spinlock_t lock;

	long remove_sequence;	/* Least-recently added (next to run) */
	long insert_sequence;	/* Next to add */

	struct list_head worklist;
	wait_queue_head_t more_work;
	wait_queue_head_t work_done;

	struct workqueue_struct *wq;
	task_t *thread;

	int run_depth;		/* Detect run_workqueue() recursion depth */
} ____cacheline_aligned;

/*
 * The externally visible workqueue abstraction is an array of
 * per-CPU workqueues:
 */
struct workqueue_struct {
	struct cpu_workqueue_struct cpu_wq[NR_CPUS];
	const char *name;
	struct list_head list; 	/* Empty if single thread */
};

/* Preempt must be disabled. */
static void __acte4_queue_work(struct cpu_workqueue_struct *cwq,
			 struct work_struct *work)
{
	unsigned long flags;
	pd3_debugk("\n");
	spin_lock_irqsave(&cwq->lock, flags);
	work->wq_data = cwq;
	list_add_tail(&work->entry, &cwq->worklist);
	cwq->insert_sequence++;
	wake_up(&cwq->more_work);
	spin_unlock_irqrestore(&cwq->lock, flags);
}

/*
 * Queue work on a workqueue. Return non-zero if it was successfully
 * added.
 *
 * We queue the work to the CPU it was submitted, but there is no
 * guarantee that it will be processed by that CPU.
 */
static inline int acte4_queue_work(struct workqueue_struct *wq, struct work_struct *work, int cpu)
{
	int ret = 0;
	pd3_debugk("\n");
	if (!test_and_set_bit(0, &work->pending)) {
		BUG_ON(!list_empty(&work->entry));
		__acte4_queue_work(wq->cpu_wq + cpu, work);
		ret = 1;
	}
	return ret;
}

#endif

static int debug=0;
static int timingcable;
static int highestorder;
static int t1e1override = -1;
static int j1mode = 0;
static int loopback = 0;
static int alarmdebounce = 0;

/* Enabling bursting can more efficiently utilize PCI bus bandwidth, but
   can also cause PCI bus starvation, especially in combination with other
   aggressive cards.  Please note that burst mode has no effect on CPU
   utilization / max number of calls / etc. */
static int noburst = 1;
static int debugslips = 0;
static int polling = 0;

#ifdef FANCY_ALARM
static int altab[] = {
0, 0, 0, 1, 2, 3, 4, 6, 8, 9, 11, 13, 16, 18, 20, 22, 24, 25, 27, 28, 29, 30, 31, 31, 32, 31, 31, 30, 29, 28, 27, 25, 23, 22, 20, 18, 16, 13, 11, 9, 8, 6, 4, 3, 2, 1, 0, 0, 
};
#endif

#define MAX_SPANS 16

#define FLAG_STARTED (1 << 0)
#define FLAG_NMF (1 << 1)
#define FLAG_SENDINGYELLOW (1 << 2)


#define	TYPE_T1	1		/* is a T1 card */
#define	TYPE_E1	2		/* is an E1 card */
#define TYPE_J1 3		/* is a running J1 */


struct devtype {
	char *desc;
	unsigned int flags;
};


static struct devtype acte1xxbnc = { "Wildcard Intelbras PXE100", 0 };
static struct devtype acte1xxrj  = { "Wildcard Intelbras PXE145", 0 };
static struct devtype acte2xxbnc = { "Wildcard Intelbras PXE200", 0 };
static struct devtype acte2xxrj  = { "Wildcard Intelbras PXE245", 0 };
static struct devtype acte4xxbnc = { "Wildcard Intelbras PXE400", 0 };
static struct devtype acte4xxrj  = { "Wildcard Intelbras PXE445", 0 };

struct ac4;

struct acte4_span {
	struct ac4 *owner;
	unsigned int *writechunk;					/* Double-word aligned write memory */
	unsigned int *readchunk;					/* Double-word aligned read memory */
	int spantype;		/* card type, T1 or E1 or J1 */
	int sync;
	int psync;
	int alarmtimer;
	int redalarms;
	int notclear;
	int alarmcount;
	int spanflags;
	int syncpos;
	int e1check;			/* E1 check */
	int reload_cas;
	unsigned char casbuf[15];
	unsigned int slipcount;
	struct zt_span span;
	unsigned char txsigs[16];	/* Transmit sigs */
	int loopupcnt;
	int loopdowncnt;
	unsigned char ec_chunk1[31][ZT_CHUNKSIZE]; /* first EC chunk buffer */
	unsigned char ec_chunk2[31][ZT_CHUNKSIZE]; /* second EC chunk buffer */
	int irqmisses;
#ifdef ENABLE_WORKQUEUES
	struct work_struct swork;
#endif	
	struct zt_chan chans[0];		/* Individual channels */
};

struct ac4 {
	/* This structure exists one per card */
	struct pci_dev *dev;		/* Pointer to PCI device */
	unsigned int intcount;
	int flag_1st_irq;	
	int num;			/* Which card we are */
	int t1e1;			/* T1/E1 select pins */
	int fpgaver;		/* version of FPGA */
	int	hwid;			/* hardware ID */
	char proc_filename[64];
	unsigned int write_key;
	int globalconfig;	/* Whether global setup has been done */
	int syncsrc;			/* active sync source */
	struct acte4_span *tspans[4];	/* Individual spans */
	int numspans;			/* Number of spans on the card */

	int blinktimer[4];
#ifdef FANCY_ALARM
	int alarmpos[4];
#endif
	int irq;			/* IRQ used by device */
	int order;			/* Order */
	int flags;			/* Device flags */
	int master;				/* Are we master */
	int ledreg;				/* LED Register */
	int stopdma;			/* Set to stop DMA */
	unsigned int dmactrl;
	int e1recover;			/* E1 recovery timer */
	dma_addr_t 	readdma;
	dma_addr_t	writedma;
	unsigned long memaddr;		/* Base address of card */
	unsigned long memlen;
	volatile unsigned int *membase;	/* Base address of card */
	volatile unsigned int *pci_vm_addr; /*acesso aos registros do PLX */
	int spansstarted;		/* number of spans started */
	/* spinlock_t lock; */		/* lock context */
	spinlock_t reglock;		/* lock register access */
	volatile unsigned int *writechunk;					/* Double-word aligned write memory */
	volatile unsigned int *readchunk;					/* Double-word aligned read memory */
	unsigned short canary;
#ifdef ENABLE_WORKQUEUES
	atomic_t worklist;
	struct workqueue_struct *workq;
#else
#ifdef ENABLE_TASKLETS
	int taskletrun;
	int taskletsched;
	int taskletpending;
	int taskletexec;
	int txerrors;
	struct tasklet_struct acte4_tlet;
#endif
#endif
	unsigned int passno;	/* number of interrupt passes */
	char *variety;
	int last0;		/* for detecting double-missed IRQ */
	int checktiming;	/* Set >0 to cause the timing source to be checked */
};


static void __set_clear(struct ac4 *wc, int span);
static int acte4_startup(struct zt_span *span);
static int acte4_shutdown(struct zt_span *span);
static int acte4_rbsbits(struct zt_chan *chan, int bits);
static int acte4_maint(struct zt_span *span, int cmd);
static int acte4_ioctl(struct zt_chan *chan, unsigned int cmd, unsigned long data);
static void __acte4_set_timing_source(struct ac4 *wc, int unit);
static void __acte4_check_alarms(struct ac4 *wc, int span);
static void __acte4_check_sigbits(struct ac4 *wc, int span);


#define AC_ACTIVATE	(1 << 12)

#define AC_OFF	(0)
#define AC_ON	(1)

#define MAX_AC4_CARDS 64

#ifdef ENABLE_TASKLETS
static void acte4_tasklet(unsigned long data);
#endif

static struct ac4 *cards[MAX_AC4_CARDS];

//#define pd3_debugk(fmt,args...) printk("pd3 -> %s: "fmt, __PRETTY_FUNCTION__, ##args)
#define pd3_debugk(fmt,args...) 

//#define TIMER_DEBUG	1

#ifdef TIMER_DEBUG
struct timer_list acte4xx_opt_timer;
unsigned int delay = 1000;
module_param(delay, uint, S_IRUGO);
#endif

#define PCI_DEVICE_ID_PLX_9030	0x9030

unsigned int acte4xx_get_serial(struct pci_dev *dev, struct ac4 *wc);
unsigned int acte4xx_e2p_date(struct pci_dev *dev, struct ac4 *wc);


static inline void __acte4_set_led(struct ac4 *wc, int span, int color)
{
	wc->ledreg &= ~(AC_ON << span);
	wc->ledreg |= (color << span);
	*(wc->membase+AC_LEDS_REG) &= ~0x000000FF;
	*(wc->membase+AC_LEDS_REG) |= ((wc->ledreg)&0xFF);
}

static inline void acte4_activate(struct ac4 *wc)
{
	wc->ledreg |= AC_ACTIVATE;
}



static void __set_clear(struct ac4 *wc, int span)
{
	int i,j;
	int oldnotclear;
	unsigned short val=0;
	struct acte4_span *ts = wc->tspans[span];
	pd3_debugk("\n");
	oldnotclear = ts->notclear;
	if (ts->spantype == TYPE_T1) {
		for (i=0;i<24;i++) {
			j = (i/8);
			if (ts->span.chans[i].flags & ZT_FLAG_CLEAR) {
				val |= 1 << (7 - (i % 8));
				ts->notclear &= ~(1 << i);
			} else
				ts->notclear |= (1 << i);
			if ((i % 8)==7) {
				val = 0;
			}
		}
	} else {
		for (i=0;i<31;i++) {
			if (ts->span.chans[i].flags & ZT_FLAG_CLEAR)
				ts->notclear &= ~(1 << i);
			else 
				ts->notclear |= (1 << i);
		}
	}
}


#ifdef CONFIG_PROC_FS
int acte4_read_proc(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	struct ac4 *wc = (struct ac4 *)data;
    int len = 0;
	
	if(!wc) {
	    *eof = 1;
	    return len;
	}

	len += sprintf(buf+len, "Intelbras PXE%d%02d Card %d\n", wc->numspans, (wc->hwid&0x01)?00:45, wc->num+1);
	len += sprintf(buf+len, "Firmware Version: %x.%x\n", (wc->fpgaver)>>8, (wc->fpgaver)&0xff);
	len += sprintf(buf+len, "Serial Number: %d\n", acte4xx_get_serial(wc->dev, wc));
	len += sprintf(buf+len, "Key In: 0x%04x\n", wc->write_key);
	len += sprintf(buf+len, "Key Out: 0x%08x%08x\n", *(wc->membase + AC_KEY_HIGH_REG), *(wc->membase + AC_KEY_LOW_REG));

    *eof = 1;
    return len;
}

int acte4_write_proc(struct file *file, const char *buffer, unsigned long count, void *data)
{
	#define KEY_STRING_SIZE	8
	struct ac4 *wc = (struct ac4 *)data;	
	char key_string[KEY_STRING_SIZE];
	unsigned int serial;

	if (copy_from_user(key_string, buffer, KEY_STRING_SIZE)) {
		return -EFAULT;
	}
	key_string[6] = '\0';
	
	serial = acte4xx_get_serial(wc->dev, wc);
	sscanf(key_string, "%04x", &wc->write_key);
	//printk("acte4: using key = 0x%04x in card with SN = %d\n", wc->write_key, serial);
	
	*(wc->membase + AC_USERRAND_REG) = AC_USERRAND_ENABLE | wc->write_key;
	*(wc->membase + AC_SERIAL_REG) = serial;
	udelay(20);
	
	return count;
}
#endif



static int acte4_ioctl(struct zt_chan *chan, unsigned int cmd, unsigned long data)
{
	struct ac4 *wc = chan->pvt;
	struct acte4_regs regs;
	int span = 0;
	int alarms = 0;
	unsigned char c, e1_cfg;

	switch(cmd) {
		case AC4_GET_ALARMS:
			if (copy_from_user(&span, (int *)data, sizeof(int)))
				return -EFAULT;
			// span starts in zero
			span--;
		        /* le status e configuracao do E1 */
		        c = ((*(wc->membase+AC_E1_STATUS_REG))>>(8*span));
		        e1_cfg = ((*(wc->membase+AC_E1_CONFIG_REG))>>(8*span));
			if( c & AC_E1_LOS_STATUS) {
				alarms = 0x01;
			} else if( c & AC_E1_AIS_STATUS) {
				alarms = 0x02;
			} else if(!(c & AC_E1_BFAE_STATUS)) {
				alarms = 0x04;
				if (c & AC_E1_RAI_STATUS)
					alarms |= 0x08;
				// Erro de MFA: 00 - MFA desabilitado, 01 - erro de MFA, 10 - MFA OK
				if ( (c & AC_E1_MFAE_STATUS) && (e1_cfg & AC_E1_CRCEN_CONFIG) )
					alarms |= 0x10;
				else if ( (!(c & AC_E1_MFAE_STATUS)) && (e1_cfg & AC_E1_CRCEN_CONFIG) )
					alarms |= 0x20;
				// Erro de CAS: 00 - desabilitado, 01 - erro de CAS, 10 - CAS OK
				if ( (!(c & AC_E1_CAS_STATUS)) && (e1_cfg & AC_E1_PCM30_CONFIG))
					alarms |= 0x40;
				else if ( (c & AC_E1_CAS_STATUS) && (e1_cfg & AC_E1_PCM30_CONFIG))
					alarms |= 0x80;
			}
			if(debug) printk("AC4_GET_ALARMS: span = %d, alarms = 0x%02x\n", span+1, alarms);
			if (copy_to_user((int *)data, &alarms, sizeof(int)))
				return -EFAULT;
			break;

		case AC4_GET_SLIPS:
			if (copy_from_user(&span, (int *)data, sizeof(int)))
				return -EFAULT;
			// span starts in zero
			span--;
			if((span < wc->numspans) && (span >=0))
				alarms = wc->tspans[span]->slipcount;
			if(debug) printk("AC4_GET_SLIPS: span = %d, slips = 0x%02x\n", span+1, alarms);
			if (copy_to_user((int *)data, &alarms, sizeof(int)))
				return -EFAULT;
			break;

		
		case AC4_GET_KEY:	
			if (copy_from_user(&regs, (struct acte4_regs *)data, sizeof(struct acte4_regs)))
				return -EFAULT;
			regs.serial = acte4xx_get_serial(wc->dev, wc);
			regs.date = acte4xx_e2p_date(wc->dev, wc);
			*(wc->membase + AC_USERRAND_REG) = AC_USERRAND_ENABLE | regs.key_low;
			*(wc->membase + AC_SERIAL_REG) = regs.serial;
			udelay(200);
			regs.key_high = *(wc->membase + AC_KEY_HIGH_REG);
			regs.key_low = *(wc->membase + AC_KEY_LOW_REG);
			printk("serial number = %d\n", regs.serial);
			printk("key = 0x%08x%08x\n", regs.key_high, regs.key_low);
	
			if (copy_to_user((unsigned long *)data, &regs, sizeof(struct acte4_regs)))
				return -EFAULT;

			break;
		default:
			return -ENOTTY;
	}
	return 0;
}

static int acte4_maint(struct zt_span *span, int cmd)
{
	struct acte4_span *ts = span->pvt;
	struct ac4 *wc = ts->owner;

	pd3_debugk("\n");
	if (ts->spantype == TYPE_E1) {
		switch(cmd) {
		case ZT_MAINT_NONE:
			printk("XXX Turn off local and remote loops E1 XXX\n");
			*(wc->membase+AC_E1_CONFIG_REG) &= ~(AC_E1_LOOP_CONFIG<<((span->spanno-1)*8));
			break;
		case ZT_MAINT_LOCALLOOP:
			printk("XXX Turn on local loopback E1 XXX\n");
			break;
		case ZT_MAINT_REMOTELOOP:
			printk("XXX Turn on remote loopback E1 XXX\n");
			break;
		case ZT_MAINT_LOOPUP:
			printk("XXX Turn on local loopback on E1 #%d instead of send loopup code XXX\n", span->spanno);
			*(wc->membase+AC_E1_CONFIG_REG) |= (AC_E1_LOOP_CONFIG<<((span->spanno-1)*8));
			break;
		case ZT_MAINT_LOOPDOWN:
			printk("XXX Turn on local loopback on E1 #%d instead of send loopdown code XXX\n", span->spanno);
			*(wc->membase+AC_E1_CONFIG_REG) |= (AC_E1_LOOP_CONFIG<<((span->spanno-1)*8));
			break;
		case ZT_MAINT_LOOPSTOP:
			printk("XXX Stop sending loop codes on E1 #%d XXX\n", span->spanno);
			*(wc->membase+AC_E1_CONFIG_REG) &= ~(AC_E1_LOOP_CONFIG<<((span->spanno-1)*8));			
			break;
		default:
			printk("PXE%d%02d: Unknown E1 maint command: %d\n", wc->numspans, (wc->hwid&0x01)?00:45, cmd);
			break;
		}
	} else {
		switch(cmd) {
	    case ZT_MAINT_NONE:
			printk("XXX Turn off local and remote loops T1 XXX\n");
			break;
	    case ZT_MAINT_LOCALLOOP:
			printk("XXX Turn on local loop and no remote loop XXX\n");
			break;
	    case ZT_MAINT_REMOTELOOP:
			printk("XXX Turn on remote loopup XXX\n");
			break;
	    case ZT_MAINT_LOOPUP:
			break;
	    case ZT_MAINT_LOOPDOWN:
			break;
	    case ZT_MAINT_LOOPSTOP:
			break;
	    default:
			printk("PXE%d%02d: Unknown T1 maint command: %d\n", wc->numspans, (wc->hwid&0x01)?00:45, cmd);
			break;
	   }
    }
	return 0;
}

static int acte4_rbsbits(struct zt_chan *chan, int bits)
{
	u_char m,c;
	int k,n,b;
	struct ac4 *wc = chan->pvt;
	struct acte4_span *ts = wc->tspans[chan->span->offset];
	unsigned long flags;
	volatile unsigned int *writecas = (wc->membase+AC_CAS_BASE);
	unsigned int allspansbits;
	
	//pd3_debugk("chan->channo = %d, int bits = 0x%08x\n", chan->channo, bits);
	if(debug & DEBUG_RBS) printk("Setting bits to %d on channel %s\n", bits, chan->name);
	spin_lock_irqsave(&wc->reglock, flags);	
	k = chan->span->offset;
	if (ts->spantype == TYPE_E1) { /* do it E1 way */
		if (chan->chanpos == 16) {
			spin_unlock_irqrestore(&wc->reglock, flags);
			return 0;
		}
		n = chan->chanpos - 1;
		if (chan->chanpos > 15) n--;
		b = (n % 15);
		c = ts->txsigs[b];
		m = (n / 15) << 2; /* nibble selector */
		c &= (0xf << m); /* keep the other nibble */
		c |= (bits & 0xf) << (4 - m); /* put our new nibble here */
		ts->txsigs[b] = c;
		/* monta a word de 32 bits com informacao de todos os spans */
		allspansbits =  wc->tspans[0]->txsigs[b];
		if (wc->numspans > 1) {
			allspansbits |=	(wc->tspans[1]->txsigs[b] << 8);
		}
		if (wc->numspans == 4) {
			allspansbits |=	(wc->tspans[2]->txsigs[b] << 16) |
							(wc->tspans[3]->txsigs[b] << 24);		
		}
		/* output them to the chip */
		writecas[b] = allspansbits;
		pd3_debugk("escrito 0x%08x para ser transmitido pelo CAS (b = %d)\n", allspansbits, b);
	} else if (ts->span.lineconfig & ZT_CONFIG_D4) {
		n = chan->chanpos - 1;
		b = (n/4);
		c = ts->txsigs[b];
		m = ((3 - (n % 4)) << 1); /* nibble selector */
		c &= ~(0x3 << m); /* keep the other nibble */
		c |= ((bits >> 2) & 0x3) << m; /* put our new nibble here */
		ts->txsigs[b] = c;
		  /* output them to the chip */
		//__acte4_out( ... ); 
	} else if (ts->span.lineconfig & ZT_CONFIG_ESF) {
		n = chan->chanpos - 1;
		b = (n/2);
		c = ts->txsigs[b];
		m = ((n % 2) << 2); /* nibble selector */
		c &= (0xf << m); /* keep the other nibble */
		c |= (bits & 0xf) << (4 - m); /* put our new nibble here */
		ts->txsigs[b] = c;
		  /* output them to the chip */
		//__acte4_out( ... ); 
	} 
	spin_unlock_irqrestore(&wc->reglock, flags);
	if (debug & DEBUG_RBS)
		printk("Finished setting RBS bits\n");
	return 0;
}

static int acte4_shutdown(struct zt_span *span)
{
	int tspan;
	int wasrunning;
	unsigned long flags;
	struct acte4_span *ts = span->pvt;
	struct ac4 *wc = ts->owner;
	
	pd3_debugk("\n");
	tspan = span->offset + 1;
	if (tspan < 0) {
		printk("PXE%dXX: Span '%d' isn't us?\n", wc->numspans, span->spanno);
		return -1;
	}

	spin_lock_irqsave(&wc->reglock, flags);
	wasrunning = span->flags & ZT_FLAG_RUNNING;

	span->flags &= ~ZT_FLAG_RUNNING;
	if (wasrunning)
		wc->spansstarted--;
	__acte4_set_led(wc, span->offset, AC_OFF);
	if (((wc->numspans == 4) && 
	    (!(wc->tspans[0]->span.flags & ZT_FLAG_RUNNING)) &&
	    (!(wc->tspans[1]->span.flags & ZT_FLAG_RUNNING)) &&
	    (!(wc->tspans[2]->span.flags & ZT_FLAG_RUNNING)) &&
	    (!(wc->tspans[3]->span.flags & ZT_FLAG_RUNNING)))
	    			|| 
	    ((wc->numspans == 2) && 
	    (!(wc->tspans[0]->span.flags & ZT_FLAG_RUNNING)) &&
	    (!(wc->tspans[1]->span.flags & ZT_FLAG_RUNNING)))
	    			|| 
	    ((wc->numspans == 1) && 
	    (!(wc->tspans[0]->span.flags & ZT_FLAG_RUNNING)))) {
		/* No longer in use, disable interrupts */
		printk("PXE%d%02d: Disabling interrupts since there are no active spans\n", wc->numspans, (wc->hwid&0x01)?00:45);
		wc->stopdma = 1;
	} else wc->checktiming = 1;
	spin_unlock_irqrestore(&wc->reglock, flags);
	if (debug & DEBUG_MAIN)
		printk("Span %d (%s) shutdown\n", span->spanno, span->name);
	return 0;
}

static int acte4_spanconfig(struct zt_span *span, struct zt_lineconfig *lc)
{
	int i;
	struct acte4_span *ts = span->pvt;
	struct ac4 *wc = ts->owner;
	unsigned int val;
	
	pd3_debugk("\n");
	printk("About to enter spanconfig!\n");
	if (debug & DEBUG_MAIN)
		printk("PXE%d%02d: Configuring span %d\n", wc->numspans, (wc->hwid&0x01)?00:45, span->spanno);
	/* XXX We assume lineconfig is okay and shouldn't XXX */	
	span->lineconfig = lc->lineconfig;
	span->txlevel = lc->lbo;
	span->rxlevel = 0;
	if (lc->sync < 0)
		lc->sync = 0;
	if (lc->sync > 4)
		lc->sync = 0;
	
	/* remove this span number from the current sync sources, if there */
	for(i = 0; i < wc->numspans; i++) {
		if (wc->tspans[i]->sync == span->spanno) {
			wc->tspans[i]->sync = 0;
			wc->tspans[i]->psync = 0;
		}
	}
	wc->tspans[span->offset]->syncpos = lc->sync;
	/* if a sync src, put it in proper place */
	if (lc->sync) {
		wc->tspans[lc->sync - 1]->sync = span->spanno;
		wc->tspans[lc->sync - 1]->psync = span->offset + 1;
	}
	wc->checktiming = 1;
	/* If we're already running, then go ahead and apply the changes */
	if (span->flags & ZT_FLAG_RUNNING)
		return acte4_startup(span);

	// Limpa contadores de slips, crc e bpv
	val = (*(wc->membase + AC_CNT_SLIP_REG));
	val = (*(wc->membase + AC_CNT_CRC_REG));
	val = (*(wc->membase + AC_CNT_CV_REG));

	printk("habilitando interrupcao!\n");
	// Nao considera as primeiras interrupcoes na soma das IRQs perdidas
	wc->flag_1st_irq = 10;
	// Enable LINTi1 in edge triggerable mode
	*(wc->pci_vm_addr+0x13) |= 0x00000103;
	// Enable PCI interrupt pin
	*(wc->pci_vm_addr+0x13) |= 0x00000040;
	// Limpa interrupcao da FPGA para forcar borda de subida na proxima
	val = *(wc->membase + AC_CLEAR_IRQ_REG);

	printk("Done with spanconfig!\n");
	return 0;
}

static int acte4_chanconfig(struct zt_chan *chan, int sigtype)
{
	int alreadyrunning;
	unsigned long flags;
	struct ac4 *wc = chan->pvt;
	
	pd3_debugk("\n");
	alreadyrunning = wc->tspans[chan->span->offset]->span.flags & ZT_FLAG_RUNNING;
	if (debug & DEBUG_MAIN) {
		if (alreadyrunning)
			printk("PXE%d%02d: Reconfigured channel %d (%s) sigtype %d\n", wc->numspans, (wc->hwid&0x01)?00:45, chan->channo, chan->name, sigtype);
		else
			printk("PXE%d%02d: Configured channel %d (%s) sigtype %d\n", wc->numspans, (wc->hwid&0x01)?00:45, chan->channo, chan->name, sigtype);
	}		
	spin_lock_irqsave(&wc->reglock, flags);	
	if (alreadyrunning)
		__set_clear(wc, chan->span->offset);
	spin_unlock_irqrestore(&wc->reglock, flags);	
	return 0;
}

static int acte4_open(struct zt_chan *chan)
{
	pd3_debugk("\n");
		
#ifndef LINUX26
	MOD_INC_USE_COUNT;
#else
	try_module_get(THIS_MODULE);
#endif	

	return 0;
}

static int acte4_close(struct zt_chan *chan)
{
	pd3_debugk("\n");
#ifndef LINUX26
	MOD_DEC_USE_COUNT;
#else
	module_put(THIS_MODULE);
#endif
	return 0;
}

static void init_spans(struct ac4 *wc)
{
	int x,y,c;
	int offset = 1;
	struct acte4_span *ts;
	pd3_debugk("\n");		
	if (!wc->t1e1)
		offset += 4;
	for (x=0;x<wc->numspans;x++) {
		ts = wc->tspans[x];
		sprintf(ts->span.name, "PXE%d%02d/%d/%d", wc->numspans, (wc->hwid&0x01)?00:45, wc->num, x + 1);
		sprintf(ts->span.desc, "Intelbras PXE%d%02d Card %d Span %d", wc->numspans, (wc->hwid&0x01)?00:45, wc->num+1, x+1);
		ts->span.spanconfig = acte4_spanconfig;
		ts->span.chanconfig = acte4_chanconfig;
		ts->span.startup = acte4_startup;
		ts->span.shutdown = acte4_shutdown;
		ts->span.rbsbits = acte4_rbsbits;
		ts->span.maint = acte4_maint;
		ts->span.open = acte4_open;
		ts->span.close  = acte4_close;
		if (ts->spantype == TYPE_T1 || ts->spantype == TYPE_J1) {
			ts->span.channels = 24;
			ts->span.deflaw = ZT_LAW_MULAW;
			/* Alteraes necessarias para o driver ficar compativel com
			 * zaptel-1.4.9 para o o uso do utilitario ztscan */
			ts->span.linecompat = ZT_CONFIG_AMI | ZT_CONFIG_B8ZS | ZT_CONFIG_D4 | ZT_CONFIG_ESF;
			ts->span.spantype = "T1";
		} else {
			ts->span.channels = 31;
			ts->span.deflaw = ZT_LAW_ALAW;
			ts->span.linecompat = ZT_CONFIG_HDB3 | ZT_CONFIG_CCS | ZT_CONFIG_CRC4;
			ts->span.spantype = "E1";
		}
		ts->span.chans = ts->chans;
		ts->span.flags = ZT_FLAG_RBS;
		/* ts->span.linecompat = ZT_CONFIG_AMI | ZT_CONFIG_B8ZS | ZT_CONFIG_D4 | ZT_CONFIG_ESF; */
		ts->span.ioctl = acte4_ioctl;
		ts->span.pvt = ts;
		ts->owner = wc;
		ts->span.offset = x;
		ts->writechunk = (void *)(wc->writechunk + x * 32 * 2);
		ts->readchunk = (void *)(wc->readchunk + x * 32 * 2);
		init_waitqueue_head(&ts->span.maintq);
		for (y=0;y<wc->tspans[x]->span.channels;y++) {
			struct zt_chan *mychans = ts->chans + y;
			sprintf(mychans->name, "PXE%d%02d/%d/%d/%d", wc->numspans, (wc->hwid&0x01)?00:45, wc->num, x + 1, y + 1);
			mychans->sigcap = ZT_SIG_EM | ZT_SIG_CLEAR | ZT_SIG_FXSLS | ZT_SIG_FXSGS | ZT_SIG_FXSKS |
									 ZT_SIG_FXOLS | ZT_SIG_FXOGS | ZT_SIG_FXOKS | ZT_SIG_CAS | ZT_SIG_EM_E1 | ZT_SIG_DACS_RBS;
			c = (x * ts->span.channels) + y;
			mychans->pvt = wc;
			mychans->chanpos = y + 1;
		}
	}
}



static void __acte4_set_timing_source(struct ac4 *wc, int unit)
{
	unsigned int timing;
	int x;

	if (unit != wc->syncsrc) {
		if ((unit > -1) && (unit < 4)) {		
			/* define fonte de clock para interface escolhida */
			timing = *(wc->membase+AC_CLKSRC_REG);
			timing &= ~AC_CLKSRC_MASK;
			timing |= unit+1;
			*(wc->membase+AC_CLKSRC_REG) = timing;
		} else {
			/* define clock para interno */
			timing = *(wc->membase+AC_CLKSRC_REG);
			timing &= ~AC_CLKSRC_MASK;
			*(wc->membase+AC_CLKSRC_REG) = timing;		
		}
		wc->syncsrc = unit;
		if ((unit < 0) || (unit > 3))
			unit = 0;
		else
			unit++;
		for (x=0;x<wc->numspans;x++)
			wc->tspans[x]->span.syncsrc = unit;
	} else {
		if (debug & DEBUG_MAIN)
			printk("PXE%d%02d: Timing source already set to %d\n", wc->numspans, (wc->hwid&0x01)?00:45, unit);
	}		
	printk("PXE%d%02d: Timing source set to %d (clksrc_reg = 0x%08x)\n", wc->numspans, (wc->hwid&0x01)?00:45, unit, *(wc->membase+AC_CLKSRC_REG));
}

static void __acte4_set_timing_source_auto(struct ac4 *wc)
{
	int x;

	wc->checktiming = 0;
	for (x=0;x<wc->numspans;x++) {
		if (wc->tspans[x]->sync) {
			if ((wc->tspans[wc->tspans[x]->psync - 1]->span.flags & ZT_FLAG_RUNNING) && 
				!(wc->tspans[wc->tspans[x]->psync - 1]->span.alarms & (ZT_ALARM_RED | ZT_ALARM_BLUE) )) {
					/* Valid timing source */
					__acte4_set_timing_source(wc, wc->tspans[x]->psync - 1);
					return;
			}
		}
	}
	__acte4_set_timing_source(wc, 4);
}

static void __acte4_configure_t1(struct ac4 *wc, int unit, int lineconfig, int txlevel)
{
	return;
}

static void __acte4_configure_e1(struct ac4 *wc, int unit, int lineconfig)
{
	char *crc4 = "";
	char *framing, *line;
	unsigned int e1s_cfg, config = 0;

	pd3_debugk("\n");	
	
	if (loopback) {
	}
		
	if (lineconfig & ZT_CONFIG_CRC4) {
		crc4 = "/CRC4";
		config |= AC_E1_CRCEN_CONFIG;
	}

	/* Configure line interface */
	if (lineconfig & ZT_CONFIG_AMI) {
		*(wc->membase+AC_LEDS_REG) |= AC_LINECODE_AMI;
		line = "AMI";
	} else {
		*(wc->membase+AC_LEDS_REG) &= ~AC_LINECODE_AMI;
		line = "HDB3";
	}
	
	if (lineconfig & ZT_CONFIG_CCS) {
		framing = "CCS";
	} else {
		framing = "CAS";
		config |= (AC_E1_CASEN_CONFIG | AC_E1_PCM30_CONFIG);
	}
	
	e1s_cfg = *(wc->membase+AC_E1_CONFIG_REG);
	e1s_cfg &= ~(0x000000ff<<(8*unit));
	e1s_cfg |= (config<<(8*unit));
	*(wc->membase+AC_E1_CONFIG_REG) = e1s_cfg;
	
	if (!polling) {
		__acte4_check_alarms(wc, unit);		
		__acte4_check_sigbits(wc, unit);
	}
	printk("PXE%d%02d: Span %d configured for %s/%s%s\n", wc->numspans, (wc->hwid&0x01)?00:45, unit + 1, framing, line, crc4);
}

static int acte4_startup(struct zt_span *span)
{
	int i;
	int tspan;
	unsigned long flags;
	int alreadyrunning;
	struct acte4_span *ts = span->pvt;
	struct ac4 *wc = ts->owner;
	pd3_debugk("\n");
	printk("About to enter startup!\n");
	tspan = span->offset + 1;
	if (tspan < 0) {
		printk("PXE%d%02d: Span '%d' isn't us?\n", wc->numspans, (wc->hwid&0x01)?00:45, span->spanno);
		return -1;
	}

	spin_lock_irqsave(&wc->reglock, flags);

	alreadyrunning = span->flags & ZT_FLAG_RUNNING;

	/* initialize the start value for the entire chunk of last ec buffer */
	for(i = 0; i < span->channels; i++)
	{
		memset(ts->ec_chunk1[i],
			ZT_LIN2X(0,&span->chans[i]),ZT_CHUNKSIZE);
		memset(ts->ec_chunk2[i],
			ZT_LIN2X(0,&span->chans[i]),ZT_CHUNKSIZE);
	}

	/* Force re-evaluation fo timing source */
//	if (timingcable)
		wc->syncsrc = -1;

	if (ts->spantype == TYPE_E1) { /* if this is an E1 card */
		__acte4_configure_e1(wc, span->offset, span->lineconfig);
	} else { /* is a T1 card */
		__acte4_configure_t1(wc, span->offset, span->lineconfig, span->txlevel);
	}

	/* Note clear channel status */
	wc->tspans[span->offset]->notclear = 0;
	__set_clear(wc, span->offset);
	
	if (!alreadyrunning) {
		span->flags |= ZT_FLAG_RUNNING;
		wc->spansstarted++;
		/* enable interrupts */
		/* Start DMA, enabling DMA interrupts on read only */
		wc->dmactrl = 1 << 29;
		wc->dmactrl |= 0xc0000003;

		if (noburst)
			wc->dmactrl |= (1 << 26);
		if (!polling) {
			__acte4_check_alarms(wc, span->offset);		
			__acte4_check_sigbits(wc, span->offset);		
		}
	}

	spin_unlock_irqrestore(&wc->reglock, flags);
	if (wc->tspans[0]->sync == span->spanno) printk("SPAN %d: Primary Sync Source\n",span->spanno);
	if (wc->numspans > 1) {	
		if (wc->tspans[1]->sync == span->spanno) printk("SPAN %d: Secondary Sync Source\n",span->spanno);
	}
	if (wc->numspans == 4) {
		if (wc->tspans[2]->sync == span->spanno) printk("SPAN %d: Tertiary Sync Source\n",span->spanno);
		if (wc->tspans[3]->sync == span->spanno) printk("SPAN %d: Quaternary Sync Source\n",span->spanno);
	}

	printk("Completed startup!\n");
	return 0;
}


static void acte4_receiveprep(struct ac4 *wc)
{
	volatile unsigned int *readchunk;
	int x,y,z;
	unsigned int tmp;
	unsigned char *byte_ptr = (unsigned char *)&tmp;

	readchunk = (wc->membase + (AC_DATA_BASE + 1));
	for (x=0;x<ZT_CHUNKSIZE;x++) {
		for (z=0;z<31;z++) {
			// All E1 channels
			tmp = readchunk[z];
			wc->tspans[0]->span.chans[z].readchunk[x] = byte_ptr[0]; // tmp & 0xff;
			if (wc->numspans > 1) {
				wc->tspans[1]->span.chans[z].readchunk[x] = byte_ptr[1]; // (tmp & 0xff00) >> 8;
			}
			if (wc->numspans == 4) {
				wc->tspans[2]->span.chans[z].readchunk[x] = byte_ptr[2]; // (tmp & 0xff0000) >> 16;
				wc->tspans[3]->span.chans[z].readchunk[x] = byte_ptr[3]; // tmp >> 24;
			}
		}
		// Advance pointer by 4 TDM frame lengths
		readchunk += 32;
	}


	for (x=0;x<wc->numspans;x++) {
		if (wc->tspans[x]->span.flags & ZT_FLAG_RUNNING) {
			for (y=0;y<wc->tspans[x]->span.channels;y++) {
				/* Echo cancel double buffered data */
				zt_ec_chunk(&wc->tspans[x]->span.chans[y], 
				    wc->tspans[x]->span.chans[y].readchunk, 
					wc->tspans[x]->ec_chunk2[y]);
				memcpy(wc->tspans[x]->ec_chunk2[y],wc->tspans[x]->ec_chunk1[y],
					ZT_CHUNKSIZE);
				memcpy(wc->tspans[x]->ec_chunk1[y],
					wc->tspans[x]->span.chans[y].writechunk,
						ZT_CHUNKSIZE);
			}
			zt_receive(&wc->tspans[x]->span);
		}
	}
}


#if (ZT_CHUNKSIZE != 8)
#error Sorry, nextgen does not support chunksize != 8
#endif

/*static inline void __receive_span(struct acte4_span *ts)
{

#ifdef ENABLE_PREFETCH
	prefetch((void *)(ts->readchunk));
	prefetch((void *)(ts->writechunk));
	prefetch((void *)(ts->readchunk + 8));
	prefetch((void *)(ts->writechunk + 8));
	prefetch((void *)(ts->readchunk + 16));
	prefetch((void *)(ts->writechunk + 16));
	prefetch((void *)(ts->readchunk + 24));
	prefetch((void *)(ts->writechunk + 24));
	prefetch((void *)(ts->readchunk + 32));
	prefetch((void *)(ts->writechunk + 32));
	prefetch((void *)(ts->readchunk + 40));
	prefetch((void *)(ts->writechunk + 40));
	prefetch((void *)(ts->readchunk + 48));
	prefetch((void *)(ts->writechunk + 48));
	prefetch((void *)(ts->readchunk + 56));
	prefetch((void *)(ts->writechunk + 56));
#endif

	zt_ec_span(&ts->span);
	zt_receive(&ts->span);
}

static inline void __transmit_span(struct acte4_span *ts)
{
	pd3_debugk("\n");
	zt_transmit(&ts->span);
}
*/

#ifdef ENABLE_WORKQUEUES
static void workq_handlespan(void *data)
{
	struct acte4_span *ts = data;
	struct ac4 *wc = ts->owner;

//	__receive_span(ts);
//	__transmit_span(ts);
	atomic_dec(&wc->worklist);
	atomic_read(&wc->worklist);

}
#endif


static void acte4_transmitprep(struct ac4 *wc)
{
	volatile unsigned int *writechunk;
	int x,y,z;
	unsigned int tmp;	

	for (y=0;y<wc->numspans;y++) {
		if (wc->tspans[y]->span.flags & ZT_FLAG_RUNNING) 
			zt_transmit(&wc->tspans[y]->span);
	}
	
	writechunk = (wc->membase+(AC_DATA_BASE+1));
	
	for (x=0;x<ZT_CHUNKSIZE;x++) {
		// Once per chunk
		for (z=0;z<31;z++) {
			// All E1 channels
			tmp = (wc->tspans[0]->span.chans[z].writechunk[x]);
			if (wc->numspans > 1) {
			  	tmp |= (wc->tspans[1]->span.chans[z].writechunk[x] << 8);
			}
			if (wc->numspans == 4) {
				tmp |=  (wc->tspans[2]->span.chans[z].writechunk[x] << 16) |
				 		(wc->tspans[3]->span.chans[z].writechunk[x] << 24);
			}
			writechunk[z] = tmp;
		}
		// Advance pointer by 4 TDM frame lengths
		writechunk += 32;
	}

}

/*static void acte4_tdm_loop(struct ac4 *wc)
{
	volatile unsigned int *buf_ptr;
	int x,z;
	unsigned int tmp;	
	
	buf_ptr = (wc->membase+AC_DATA_BASE);
	
	for (x=0;x<ZT_CHUNKSIZE;x++) {
		// Once per chunk
		for (z=0;z<32;z++) {
			tmp = buf_ptr[z];
			buf_ptr[z] = tmp;
		}
		buf_ptr += 32;
	}

}
*/
static void __acte4_check_sigbits(struct ac4 *wc, int span)
{
	int a,i,rxs;
	struct acte4_span *ts = wc->tspans[span];
	volatile unsigned int *readcas = (wc->membase+AC_CAS_BASE);

//	if (debug & DEBUG_RBS)
//		printk("Checking sigbits on span %d\n", span + 1);

	if (!(ts->span.flags & ZT_FLAG_RUNNING))
		return;

	// se span estiver com alarme RED ou BLUE...
	if( (ts->span.alarms & ZT_ALARM_RED) || (ts->span.alarms & ZT_ALARM_BLUE) ) {
		ts->reload_cas = 4;
	} else if(ts->reload_cas > 0) {
		// da mais um tempo para framer recuperar e enviar bits de CAS validos
		ts->reload_cas--;
	}
	
	if (ts->spantype == TYPE_E1) {
		for (i = 0; i < 15; i++) {
		
			// Se estamos em alarme ou recuperando de um entao mascara os bits para "1101" (bloqueado)
			if(ts->reload_cas) {
				a = 0xdd;
			} else {
				a = (int) ts->casbuf[i];			
			}
			ts->casbuf[i] = (unsigned char) (readcas[i] >> (8*span))&0xff;

			/* Get high channel in low bits */
			rxs = (a & 0xf);
			if (!(ts->span.chans[i+16].sig & ZT_SIG_CLEAR)) {
				if (ts->span.chans[i+16].rxsig != rxs) {
					pd3_debugk("CAS no canal %d mudou de 0x%02x para 0x%02x\n", i+16, ts->span.chans[i+16].rxsig, rxs);
					zt_rbsbits(&ts->span.chans[i+16], rxs);
				}
			}
			rxs = (a >> 4) & 0xf;
			if (!(ts->span.chans[i].sig & ZT_SIG_CLEAR)) {
				if (ts->span.chans[i].rxsig != rxs) {
					pd3_debugk("CAS no canal %d mudou de 0x%02x para 0x%02x\n", i, ts->span.chans[i].rxsig, rxs);
					zt_rbsbits(&ts->span.chans[i], rxs);
				}
			}
		}
	} else if (ts->span.lineconfig & ZT_CONFIG_D4) {
		for (i = 0; i < 24; i+=4) {
			a = 0;
			/* Get high channel in low bits */
			rxs = (a & 0x3) << 2;
			if (!(ts->span.chans[i+3].sig & ZT_SIG_CLEAR)) {
				if (ts->span.chans[i+3].rxsig != rxs)
					zt_rbsbits(&ts->span.chans[i+3], rxs);
			}
			rxs = (a & 0xc);
			if (!(ts->span.chans[i+2].sig & ZT_SIG_CLEAR)) {
				if (ts->span.chans[i+2].rxsig != rxs)
					zt_rbsbits(&ts->span.chans[i+2], rxs);
			}
			rxs = (a >> 2) & 0xc;
			if (!(ts->span.chans[i+1].sig & ZT_SIG_CLEAR)) {
				if (ts->span.chans[i+1].rxsig != rxs)
					zt_rbsbits(&ts->span.chans[i+1], rxs);
			}
			rxs = (a >> 4) & 0xc;
			if (!(ts->span.chans[i].sig & ZT_SIG_CLEAR)) {
				if (ts->span.chans[i].rxsig != rxs)
					zt_rbsbits(&ts->span.chans[i], rxs);
			}
		}
	} else {
		for (i = 0; i < 24; i+=2) {
			a = 0;
			/* Get high channel in low bits */
			rxs = (a & 0xf);
			if (!(ts->span.chans[i+1].sig & ZT_SIG_CLEAR)) {
				/* XXX Not really reset on every trans! XXX */
				if (ts->span.chans[i+1].rxsig != rxs) {
					zt_rbsbits(&ts->span.chans[i+1], rxs);
				}
			}
			rxs = (a >> 4) & 0xf;
			if (!(ts->span.chans[i].sig & ZT_SIG_CLEAR)) {
				/* XXX Not really reset on every trans! XXX */
				if (ts->span.chans[i].rxsig != rxs) {
					zt_rbsbits(&ts->span.chans[i], rxs);
				}
			}
		}
	}
}

static void __acte4_check_alarms(struct ac4 *wc, int span)
{
	unsigned char c;
	int alarms;
	int x,j;
	struct acte4_span *ts = wc->tspans[span];
	unsigned int e1_cfg;
				
	if (!(ts->span.flags & ZT_FLAG_RUNNING))
		return;

	/* Assume no alarms */
	alarms = ZT_ALARM_NONE;

	/* And consider only carrier alarms */
	ts->span.alarms &= (ZT_ALARM_RED | ZT_ALARM_BLUE | ZT_ALARM_NOTOPEN);

	/* le status e configuracao do E1 */
	c = ((*(wc->membase+AC_E1_STATUS_REG))>>(8*span));
	e1_cfg = ((*(wc->membase+AC_E1_CONFIG_REG))>>(8*span));
	
	if (ts->span.lineconfig & ZT_CONFIG_NOTOPEN) {
		for (x=0,j=0;x < ts->span.channels;x++)
			if ((ts->span.chans[x].flags & ZT_FLAG_OPEN) ||
			    (ts->span.chans[x].flags & ZT_FLAG_NETDEV))
				j++;
		if (!j)
			alarms |= ZT_ALARM_NOTOPEN;
	}
	
	if ((c & AC_E1_LOS_STATUS)||(c & AC_E1_BFAE_STATUS)||(c & AC_E1_AIS_STATUS)) { 
		if (ts->alarmcount >= alarmdebounce) 
			alarms |= ZT_ALARM_RED;
		else
			ts->alarmcount++;
	} else
		ts->alarmcount = 0;
		
	if ( c & AC_E1_MFAE_STATUS )
		alarms |= ZT_ALARM_BLUE;

	if ( (!(c & AC_E1_CAS_STATUS)) && (e1_cfg & AC_E1_PCM30_CONFIG))
		alarms |= ZT_ALARM_BLUE;

	if (((!ts->span.alarms) && alarms) || 
	    (ts->span.alarms && (!alarms))) 
		wc->checktiming = 1;

	/* Keep track of recovering */
	if ((!alarms) && ts->span.alarms) 
		ts->alarmtimer = ZT_ALARMSETTLE_TIME;
	if (ts->alarmtimer)
		alarms |= ZT_ALARM_RECOVER;


	// If receiving alarms, go into Yellow alarm state
	if (alarms && !(ts->spanflags & FLAG_SENDINGYELLOW)) {
		printk("Setting yellow alarm on span %d\n", span + 1);
		e1_cfg = *(wc->membase+AC_E1_CONFIG_REG);
		e1_cfg |= (AC_E1_RAI_CONFIG<<(8*span));
		*(wc->membase+AC_E1_CONFIG_REG) = e1_cfg;
		ts->spanflags |= FLAG_SENDINGYELLOW;
	} else if ((!alarms) && (ts->spanflags & FLAG_SENDINGYELLOW)) {
		printk("Clearing yellow alarm on span %d\n", span + 1);
		e1_cfg = *(wc->membase+AC_E1_CONFIG_REG);
		e1_cfg &= ~(AC_E1_RAI_CONFIG<<(8*span));
		*(wc->membase+AC_E1_CONFIG_REG) = e1_cfg;
		ts->spanflags &= ~FLAG_SENDINGYELLOW;
	}

	// Re-check the timing source when we enter/leave alarm, not withstanding yellow alarm
	if (c & AC_E1_RAI_STATUS)
		alarms |= ZT_ALARM_YELLOW;

	if (ts->span.mainttimer || ts->span.maintstat) 
		alarms |= ZT_ALARM_LOOPBACK;
		
	ts->span.alarms = alarms;
	zt_alarm_notify(&ts->span);
}

static void __acte4_do_counters(struct ac4 *wc)
{
	int span;

	for (span=0;span<wc->numspans;span++) {
		struct acte4_span *ts = wc->tspans[span];
		int docheck=0;
		if (ts->loopupcnt || ts->loopdowncnt)
			docheck++;
		if (ts->alarmtimer) {
			if (!--ts->alarmtimer) {
				docheck++;
				ts->span.alarms &= ~(ZT_ALARM_RECOVER);
			}
		}
		if (docheck) {
			if (!polling)
				__acte4_check_alarms(wc, span);
			zt_alarm_notify(&ts->span);
		}
	}
}

static inline void __handle_leds(struct ac4 *wc)
{
	int x, span_status;
	#define MAX_BLINKTIMER	0x14

	for (x=0;x<wc->numspans;x++) {
		struct acte4_span *ts = wc->tspans[x];
		/* le status do E1 (para avaliar LOS) */
		span_status = ((*(wc->membase+AC_E1_STATUS_REG))>>(8*x));
		if (ts->span.flags & ZT_FLAG_RUNNING) {
			if(span_status&AC_E1_LOS_STATUS) {
				if (wc->blinktimer[x] >= (altab[wc->alarmpos[x]] /*>> 1*/)) {
					__acte4_set_led(wc, x, AC_ON);
				}
				if (wc->blinktimer[x] >= (MAX_BLINKTIMER-1)) {
					__acte4_set_led(wc, x, AC_OFF);
				}
				wc->blinktimer[x] += 1;			
			} else if (ts->span.alarms & (ZT_ALARM_RED | ZT_ALARM_BLUE)) {
				if (wc->blinktimer[x] >= (altab[wc->alarmpos[x]] /*>> 1*/)) {
					__acte4_set_led(wc, x, AC_ON);
				}
				if (wc->blinktimer[x] >= (MAX_BLINKTIMER-2)) {
					__acte4_set_led(wc, x, AC_OFF);
				}
				wc->blinktimer[x] += 3;				
			} /*else if (ts->span.alarms & ZT_ALARM_YELLOW) {
				// Yellow Alarm
				__acte4_set_led(wc, x, AC_ON);
			} else if (ts->span.mainttimer || ts->span.maintstat) {

				if (wc->blinktimer == (altab[wc->alarmpos] >> 1)) {
					__acte4_set_led(wc, x, AC_GREEN);
				}
				if (wc->blinktimer == 0xf) {
					__acte4_set_led(wc, x, AC_OFF);
				}

			} */else {
				/* No Alarm */
				__acte4_set_led(wc, x, AC_ON);
			}
		}	else
				__acte4_set_led(wc, x, AC_OFF);

		if (wc->blinktimer[x] > MAX_BLINKTIMER) {
			wc->blinktimer[x] = 0;
			wc->alarmpos[x]++;
			if (wc->alarmpos[x] >= (sizeof(altab) / sizeof(altab[0])))
				wc->alarmpos[x] = 0;
		}
	
	}


	
}


ZAP_IRQ_HANDLER(acte4_interrupt)
{
	struct ac4 *wc = dev_id;
	unsigned long flags;
	int x;
	static unsigned int val, cfg;
	unsigned int cnt_irq_misses;
	static unsigned int cnt_tmp;
	
	/* retorna se interrupcao nao foi habilitada */
	cfg = *(wc->pci_vm_addr+0x13);
	if((cfg&0x0040) == 0) {
#ifdef LINUX26
		return IRQ_NONE;
#else
		return;
#endif
	}
       /* retorna se interrupcao nao esta ativa */
	if((cfg&0x0004) == 0) {
#ifdef LINUX26
		return IRQ_NONE;
#else
		return;
#endif
	}
	/* ack da interrupcao */
	*(wc->pci_vm_addr+0x13) = (cfg|0x00000400);
	val = *(wc->membase + AC_CLEAR_IRQ_REG);

	/* conta interrupcoes perdidas */
	if (wc->flag_1st_irq > 0) {
		// nao considera as primeiras passagens pela rotina
		cnt_irq_misses = (*(wc->membase+AC_CNT_IRQ_REG));
		// so considera int. para o cartao
		if(cnt_irq_misses) {
			wc->flag_1st_irq--;
			*(wc->membase+AC_CNT_IRQ_REG)=0;
			for(x=0;x<(wc->numspans);x++)	
				wc->tspans[x]->span.irqmisses = 0;
		}
	} else {
		// neste registro da FPGA temos o numero de interrupcoes que aconteceram
		// desde o ultimo reset do contador de interrupcoes. O normal eh ler 1.
		cnt_irq_misses = (*(wc->membase+AC_CNT_IRQ_REG));
		// Se for zero significa que a interrupcao nao foi gerada pelo nosso cartao
		if(cnt_irq_misses == 0) {
			if(debug) printk("Interrupcao gerada mas nao pela FPGA?!\n");
#ifdef LINUX26
			return IRQ_NONE;
#else
			return;
#endif
		}				
		// reseta o contador
		*(wc->membase+AC_CNT_IRQ_REG)=0;		
		for(x=0;x<(wc->numspans);x++)	
			wc->tspans[x]->span.irqmisses += (cnt_irq_misses-1);
	}
			
	if (!wc->spansstarted) {
		/* Not prepped yet! */
#ifdef LINUX26
		return IRQ_NONE;
#else
		return;
#endif		
	}

	wc->intcount++;
	
#ifdef ENABLE_WORKQUEUES
	int cpus = num_online_cpus();
	atomic_set(&wc->worklist, wc->numspans);
	if (wc->tspans[0]->span.flags & ZT_FLAG_RUNNING)
		acte4_queue_work(wc->workq, &wc->tspans[0]->swork, 0);
	else
		atomic_dec(&wc->worklist);
	if (wc->numspans > 1) {
		if (wc->tspans[1]->span.flags & ZT_FLAG_RUNNING)
			acte4_queue_work(wc->workq, &wc->tspans[1]->swork, 1 % cpus);
		else
			atomic_dec(&wc->worklist);
	}
	if (wc->numspans == 4) {
		if (wc->tspans[2]->span.flags & ZT_FLAG_RUNNING)
			acte4_queue_work(wc->workq, &wc->tspans[2]->swork, 2 % cpus);
		else
			atomic_dec(&wc->worklist);
		if (wc->tspans[3]->span.flags & ZT_FLAG_RUNNING)
			acte4_queue_work(wc->workq, &wc->tspans[3]->swork, 3 % cpus);
		else
			atomic_dec(&wc->worklist);
	}
#else
	acte4_receiveprep(wc);
	acte4_transmitprep(wc);
	
//	acte4_tdm_loop(wc);
#endif	
		
	// Estatisticas a cada 128ms
	if(!(wc->intcount&0x7f)){
		cnt_tmp = (*(wc->membase + AC_CNT_CV_REG));
		for(x=0;x<(wc->numspans);x++)
			wc->tspans[x]->span.bpvcount += (cnt_tmp>>(8*x))&0xff;
		cnt_tmp = (*(wc->membase + AC_CNT_CRC_REG));
		for(x=0;x<(wc->numspans);x++) 
			wc->tspans[x]->span.crc4count += (cnt_tmp>>(8*x))&0xff;		
		cnt_tmp = (*(wc->membase + AC_CNT_SLIP_REG));
		for(x=0;x<(wc->numspans);x++) {
			if (((cnt_tmp>>(8*x))&0xff) && (!(wc->tspans[x]->span.alarms & ZT_ALARM_RED)) ){
				wc->tspans[x]->slipcount++;
				if(debug) printk("Slip detected on span %d: slipcount = %d\n", x+1, wc->tspans[x]->slipcount);
			}		
		}
	}


	spin_lock_irqsave(&wc->reglock, flags);

	__handle_leds(wc);
	
	__acte4_do_counters(wc);

	//x = wc->intcount & 15;
	x = wc->intcount & 7;
	switch(x) {
	case 0:
	case 1:
	case 2:
	case 3:
		__acte4_check_alarms(wc, x);		
		break;
	case 4:
	case 5:
	case 6:
	case 7:
		__acte4_check_sigbits(wc, x - 4);
		break;
	}

	if (wc->checktiming > 0)
		__acte4_set_timing_source_auto(wc);
		
	spin_unlock_irqrestore(&wc->reglock, flags);

#ifdef LINUX26
	return IRQ_RETVAL(1);
#endif		
}


static int __devinit acte4_launch(struct ac4 *wc)
{
	int x;
	unsigned long flags;
	pd3_debugk("\n");
	if (wc->tspans[0]->span.flags & ZT_FLAG_REGISTERED)
		return 0;
	printk("PXE%d%02d: Launching card: %d\n", wc->numspans, (wc->hwid&0x01)?00:45, wc->order);

	/* Setup serial parameters and system interface */
	for (x=0;x<4;x++) {
		//acte4_serial_setup(wc, x);
		wc->globalconfig = 1;
	}

	if (zt_register(&wc->tspans[0]->span, 0)) {
		printk(KERN_ERR "Unable to register span %s\n", wc->tspans[0]->span.name);
		return -1;
	}
	if (wc->numspans > 1) {
		if (zt_register(&wc->tspans[1]->span, 0)) {
			printk(KERN_ERR "Unable to register span %s\n", wc->tspans[1]->span.name);
			zt_unregister(&wc->tspans[0]->span);
			return -1;
		}
	}
	if (wc->numspans == 4) {
		if (zt_register(&wc->tspans[2]->span, 0)) {
			printk(KERN_ERR "Unable to register span %s\n", wc->tspans[2]->span.name);
			zt_unregister(&wc->tspans[0]->span);
			zt_unregister(&wc->tspans[1]->span);
			return -1;
		}
		if (zt_register(&wc->tspans[3]->span, 0)) {
			printk(KERN_ERR "Unable to register span %s\n", wc->tspans[3]->span.name);
			zt_unregister(&wc->tspans[0]->span);
			zt_unregister(&wc->tspans[1]->span);
			zt_unregister(&wc->tspans[2]->span);
			return -1;
		}
	}
	wc->checktiming = 1;
	spin_lock_irqsave(&wc->reglock, flags);
//	__acte4_set_timing_source(wc,4);
	spin_unlock_irqrestore(&wc->reglock, flags);
#ifdef ENABLE_TASKLETS
	tasklet_init(&wc->acte4_tlet, acte4_tasklet, (unsigned long)wc);
#endif
	return 0;
}


int acte4xx_liu_reset(struct ac4 *wc)
{
        unsigned int jiffies_hold = jiffies;

        *(wc->membase+AC_LEDS_REG) |= AC_LIU_RESET_BIT;
	while(jiffies<=(jiffies_hold+2));
	*(wc->membase+AC_LEDS_REG) &= ~AC_LIU_RESET_BIT;
        return 0;
}


int acte4xx_spi_write(struct ac4 *wc, int addr, int val)
{
	unsigned int jiffies_hold = jiffies;

	addr &= 0xff;
	val = (val<<8)&0xff00;
	*(wc->membase+AC_SPI_CTL_REG) = AC_SPI_FLAG_WRITE | val |addr;
	while( (*(wc->membase+AC_SPI_STA_REG)) & AC_SPI_FLAG_BUSY) {
		if(jiffies>(jiffies_hold+1)) {
			printk("acte4xx_spi_write: write timeout (0x%08x)!\n", *(wc->membase+AC_SPI_STA_REG));
			return -1;
		}
	}
	return 0;
}

int acte4xx_spi_read(struct ac4 *wc, int addr)
{
	unsigned int jiffies_hold = jiffies;

	addr &= 0xff;
	*(wc->membase+AC_SPI_CTL_REG) = AC_SPI_FLAG_READ | addr;
	while( (*(wc->membase+AC_SPI_STA_REG)) & AC_SPI_FLAG_BUSY) {
		if(jiffies>(jiffies_hold+1)) {
			printk("acte4xx_spi_read: read timeout (0x%08x)!\n", *(wc->membase+AC_SPI_STA_REG));
			return -1;
		}
	}
	return ((*(wc->membase+AC_SPI_STA_REG)) & 0xff);
}

int acte4xx_spi_test(struct ac4 *wc)
{
	int i, res, orig;
	
	/* utiliza o reg 0x20 como registro de teste */
	orig = acte4xx_spi_read(wc, 0x20);
	
	for(i=0;i<256;i++) {
		if(acte4xx_spi_write(wc, 0x20, i)) {
			pd3_debugk("Erro escrevendo na SPI!\n");
			return -1;
		}
		res = acte4xx_spi_read(wc, 0x20);
		if(res<0) {
			pd3_debugk("Erro lendo da SPI!\n");
			return -1;
		}
		if(res != i) {
			pd3_debugk("Valor escrito diferente do valor lido!\n");
			return -1;		
		}
	}

	if(acte4xx_spi_write(wc, 0x20, orig)) {
		pd3_debugk("Erro restaurando valor do registro!\n");
		return -1;
	}

	return 0;
}

u32 acte4xx_e2prom_read(struct pci_dev *dev, u16 addr)
{
	unsigned int jiffies_hold = jiffies;
	u16 data16;
	u32 data32;
	
	data16 = addr;
	pci_write_config_word(dev, AC_E2P_PVDADDR_REG, data16);	
	while(!(data16 & AC_E2P_OPER_FLAG)) {
		pci_read_config_word(dev, AC_E2P_PVDADDR_REG, &data16);
		if(jiffies>(jiffies_hold+100)) {
			printk("acte4xx_e2prom_read: timeout!\n");
			return -1;
		}
	}
	pci_read_config_dword(dev, AC_E2P_PVDATA_REG, &data32);
	return data32;
}

unsigned int acte4xx_get_serial(struct pci_dev *dev, struct ac4 *wc)
{
	unsigned int sn = acte4xx_e2prom_read(dev, AC_E2P_ADDR_SERIAL);

	if(sn == 0xFFFFFFFF) return 0;
	else return sn;
}

unsigned int acte4xx_e2p_date(struct pci_dev *dev, struct ac4 *wc)
{
	unsigned int date = acte4xx_e2prom_read(dev, AC_E2P_ADDR_DATE);
	
	if(date == 0xFFFFFFFF) return 0;
	else return date;
}

int acte4xx_bus_test(struct ac4 *wc)
{
	int tst_result = 0;
	unsigned int val;

	*(wc->membase+AC_E1_CONFIG_REG) = 0xAAAAAAAA;
	*wc->membase = 0; // flush
	val = *(wc->membase+AC_E1_CONFIG_REG);
	if(val != 0xAAAAAAAA) {
		printk("Escrito 0xAAAAAAAA, lido 0x%08X!\n", val);
		tst_result++;
	}
	*(wc->membase+AC_E1_CONFIG_REG) = 0x55555555;
	*wc->membase = 0; // flush
        val = *(wc->membase+AC_E1_CONFIG_REG);
	if(val != 0x55555555) {
		printk("Escrito 0x55555555, lido 0x%08X!\n", val);
		tst_result++;
	}
	*(wc->membase+AC_E1_CONFIG_REG) = 0x00000000;
	*wc->membase = 0xFFFFFFFF; // flush
	val = *(wc->membase+AC_E1_CONFIG_REG);
	if(val != 0x00000000) {
		printk("Escrito 0x00000000, lido 0x%08X!\n", val);
		tst_result++;
	}	
	return tst_result;
}

#ifdef TIMER_DEBUG
void acte4xx_opt_timeout(unsigned long arg)
{
	struct pci_dev *dev = (struct pci_dev *)arg;
	struct ac4 *wc = pci_get_drvdata(dev);
	static unsigned int intcount_old = 0;
//	unsigned int led_count, val;

//	pd3_debugk("ips = %d (%d)\n", ((wc->intcount)-intcount_old), wc->intcount);
	intcount_old = wc->intcount;
	
//	pd3_debugk("wc->tspans[0]->span.chans[1].readchunk[1] = 0x%02x\n", wc->tspans[0]->span.chans[0].readchunk[1]);
//	pd3_debugk("e1s_cfg = 0x%08x\n", *(wc->membase+AC_E1_CONFIG_REG));
//	pd3_debugk("clk_cfg = 0x%08x\n", *(wc->membase+0x07));
//	pd3_debugk("e1_data = 0x%08x\n", *(wc->membase + (AC_DATA_BASE + 1)));
//	pd3_debugk("cas_data = 0x%08x\n", *(wc->membase + AC_CAS_BASE));
//	pd3_debugk("e1_status = 0x%08x\n", *(wc->membase + AC_E1_STATUS_REG));
//	pd3_debugk("leds_reg = 0x%08x\n", *(wc->membase + AC_LEDS_REG));

//	*(wc->membase+AC_SPI_CTL_REG) = 0;
//	printk("*(wc->membase+AC_SPI_STA_REG) = 0x%08x\n", *(wc->membase+AC_SPI_STA_REG));	
	
//	pd3_debugk("jiffies =  %d\n", jiffies);		

	// dispara timer novamente
	init_timer(&acte4xx_opt_timer);
	acte4xx_opt_timer.function = acte4xx_opt_timeout;
	acte4xx_opt_timer.data = arg;
	acte4xx_opt_timer.expires = jiffies + (delay/4);
	add_timer(&acte4xx_opt_timer);
	
}
#endif

			
static int acte4xx_init(struct pci_dev *dev, const struct pci_device_id *id, struct ac4 *wc)
{
	unsigned long int pci_start_hwaddr, pci_end_hwaddr, pci_len_hwaddr;
	u8 irq_num;
	u16 data;
	u32 word32;


	pd3_debugk("\n");

	/* Verifica VENDOR e DEVICE */
	pci_read_config_word(dev, PCI_VENDOR_ID, &data);
	if(data != PCI_VENDOR_ID_PLX) {
		printk("PCI_VENDOR_ID (0x%04x) diferente do esperado para o cartao (0x%04x)\n", data, PCI_VENDOR_ID_PLX);
		return -ENODEV;
	}
	pci_read_config_word(dev, PCI_DEVICE_ID, &data);
	if(data != PCI_DEVICE_ID_PLX_9030) {
		printk("PCI_DEVICE_ID (0x%04x) diferente do esperado para o cartao (0x%04x)\n", data, PCI_DEVICE_ID_PLX_9030);
		return -ENODEV;
	}

	/* imprime status e registro de comando */
	pci_read_config_word(dev, PCI_COMMAND, &data);
	//printk("registro de comando = 0x%04x\n", data);
	pci_read_config_word(dev, PCI_STATUS, &data);
	//printk("registro de status  = 0x%04x\n", data);

	/* le numero da IRQ associada */
	pci_read_config_byte(dev, PCI_INTERRUPT_LINE, &irq_num);
	printk("irq number = 0x%02x\n", irq_num);

	/* verifica e mapeia primeiro espaco de enderecamento
	   este espaco eh acessado via espaco de memoria e da acesso
	   aos registros de configuracao local do PLX */
	pci_start_hwaddr = pci_resource_start(dev, 0);
	pci_end_hwaddr = pci_resource_end(dev, 0);
	pci_len_hwaddr = pci_end_hwaddr - pci_start_hwaddr;
	printk("pci_resource_start = 0x%08lx\n", pci_start_hwaddr);
	printk("pci_resource_end = 0x%08lx\n", pci_end_hwaddr);

	/* aloca regioes de memoria */
	if (pci_request_regions(dev, "pxe"))
			printk("acte4xx: Unable to request regions\n");

	/* verifica se regiao de memoria esta livre e aloca a mesma*/
/*	if(request_mem_region(pci_start_hwaddr, pci_len_hwaddr, "e1_config_space") == NULL) {
		printk("request_mem_region falhou!!!\n");
		return -ENODEV;
	}
*/

	/* faz o ioremap da regiao alocada */
	wc->pci_vm_addr = ioremap_nocache(pci_start_hwaddr, pci_len_hwaddr);
	if(wc->pci_vm_addr == NULL) {
		printk("ioremap falhou!!!\n");
		release_mem_region(pci_start_hwaddr, pci_len_hwaddr);
		return -ENODEV;
	}

	/* Agora que temos ponteiro para os registros locais de configuracao,
		ativa address space 0 e CS0 */
	*(wc->pci_vm_addr+0x0a) = 0x4090A840;				// LAS0BRD (offset = 0x28) -> insere 1 wait state em todos os ciclos para nao se preocupar agora	
	*(wc->pci_vm_addr+0x0f) = 0x00080001;				// CS0BASE (offset = 0x3c) -> habilita CS0
	*(wc->pci_vm_addr+0x13) = 0x003F0000;
	pci_read_config_dword(dev, PCI_BASE_ADDRESS_2, &word32);
//	printk("PCI_BASE_ADDRESS_2 = 0x%08x\n", word32);
	printk("local cfg reg 0x00 = 0x%08x\n", *(wc->pci_vm_addr));
	
	return 0;
}

static int __devinit acte4_init_one(struct pci_dev *pdev, const struct pci_device_id *ent)
{
	int res;
	struct ac4 *wc;
	struct devtype *dt;
	int x,f, btest;
	int basesize;
	static int initd_ifaces=0;
#ifdef CONFIG_PROC_FS	
	static struct proc_dir_entry *rw_proc_file;
#endif


	pd3_debugk("\n");
	if(!initd_ifaces){
		memset((void *)cards,0,(sizeof(struct ac4 *))*MAX_AC4_CARDS);
		initd_ifaces=1;
	}
	
	if (pci_enable_device(pdev)) {
		res = -EIO;
	} else {

		wc = kmalloc(sizeof(struct ac4), GFP_KERNEL);
		if (wc) {
			memset(wc, 0x0, sizeof(struct ac4));
			spin_lock_init(&wc->reglock);
		
			if(acte4xx_init(pdev, ent, wc)<0) {
				kfree(wc);
				return -EIO;
			}
			
			basesize = ZT_MAX_CHUNKSIZE * 32 * 2 * 4;

			// mapeia agora segundo espaco de enderecamento que 
			// sera usado para acessar o barramento local do cartao
			wc->memaddr = pci_resource_start(pdev, 2);
			wc->memlen = pci_resource_len(pdev, 2);			
			//printk("pci_resource_start(dev, 2) = 0x%08lx\n", wc->memaddr);
			//printk("pci_resource_len(dev, 2) = 0x%08lx\n", wc->memlen);
			// verifica se regiao de memoria esta livre e aloca a mesma
/*			if(request_mem_region(wc->memaddr, wc->memlen, wc->variety) == NULL) {
				printk("request_mem_region falhou!!!\n");
				return -EIO;
			}*/
			// faz o ioremap da regiao alocada
			wc->membase = ioremap_nocache(wc->memaddr, wc->memlen);
			if(wc->membase == NULL) {
				printk("ioremap falhou!!!\n");
				release_mem_region(wc->memaddr, wc->memlen);
				kfree(wc);
				return -EIO;
			}
			
			// verifica modelo do cartao
			if(*wc->membase == ACTE_CARD_ID) {
				// executa teste preliminar de escrita/leitura no barramento
				if(acte4xx_bus_test(wc)) {
					printk("ERRO NO TESTE DE BARRAMENTO!!!\n");
					kfree(wc);
					return -EIO;
				}
				wc->fpgaver = (*(wc->membase+AC_CARD_TYPE_REG))&0xffff;
				if (wc->fpgaver > 0x0101) {
					*(wc->pci_vm_addr+0x0a) = 0x00800000;	// LAS0BRD (offset = 0x28) -> sem wait states
					printk("Using no wait states\n");
				} else {
					printk("Using 1 wait state\n");
				}
				// executa novo teste no barramento
				for(btest=0; btest<1000; btest++) {
					if(acte4xx_bus_test(wc)) {
						printk("ERRO NO TESTE DE BARRAMENTO!!! (%d)\n", btest);
                   		kfree(wc);
						return -EIO;
					}
				}
				printk("Barramento OK!\n");
				wc->numspans = (*(wc->membase+AC_CARD_TYPE_REG))>>16;
				wc->hwid = ((*(wc->membase+AC_HWCONFIG_REG))&AC_HWID_MASK)>>4;
				switch(wc->hwid) {
					case AC_HWID_1E1_RJ:
						if(wc->numspans != 1) {
							printk("Hardware ID incompativel com versao da FPGA (hw_id = 0x%02x)!!! Abortando...\n", wc->hwid);
							kfree(wc);							
							return -EIO;
						}
						dt = (struct devtype *)&acte1xxrj;
						printk("Connector type: RJ45, numspans = %d, fpgaver = %x.%x\n", wc->numspans, wc->fpgaver>>8, wc->fpgaver&0xff);											
						break;
					case AC_HWID_1E1_BNC:
						if(wc->numspans != 1) {
							printk("Hardware ID incompativel com versao da FPGA (hw_id = 0x%02x)!!! Abortando...\n", wc->hwid);
							kfree(wc);							
							return -EIO;
						}
						dt = (struct devtype *)&acte1xxbnc;
						printk("Connector type: BNC, numspans = %d, fpgaver = %x.%x\n", wc->numspans, wc->fpgaver>>8, wc->fpgaver&0xff);																	
						break;
					case AC_HWID_2E1_RJ:
						if(wc->numspans != 2) {
							printk("Hardware ID incompativel com versao da FPGA (hw_id = 0x%02x)!!! Abortando...\n", wc->hwid);
							kfree(wc);							
							return -EIO;
						}
						dt = (struct devtype *)&acte2xxrj;
						printk("Connector type: RJ45, numspans = %d, fpgaver = %x.%x\n", wc->numspans, wc->fpgaver>>8, wc->fpgaver&0xff);
						break;
					case AC_HWID_2E1_BNC:
						if(wc->numspans != 2) {
							printk("Hardware ID incompativel com versao da FPGA (hw_id = 0x%02x)!!! Abortando...\n", wc->hwid);
							kfree(wc);							
							return -EIO;
						}
						dt = (struct devtype *)&acte2xxbnc;
						printk("Connector type: BNC, numspans = %d, fpgaver = %x.%x\n", wc->numspans, wc->fpgaver>>8, wc->fpgaver&0xff);											
						break;
					case AC_HWID_4E1_RJ:
						if(wc->numspans != 4) {
							printk("Hardware ID incompativel com versao da FPGA (hw_id = 0x%02x)!!! Abortando...\n", wc->hwid);
							kfree(wc);							
							return -EIO;
						}
						dt = (struct devtype *)&acte4xxrj;
						printk("Connector type: RJ45, numspans = %d, fpgaver = %x.%x\n", wc->numspans, wc->fpgaver>>8, wc->fpgaver&0xff);
						break;
					case AC_HWID_4E1_BNC:
						if(wc->numspans != 4) {
							printk("Hardware ID incompativel com versao da FPGA (hw_id = 0x%02x)!!! Abortando...\n", wc->hwid);
							kfree(wc);							
							return -EIO;
						}
						dt = (struct devtype *)&acte4xxbnc;
						printk("Connector type: BNC, numspans = %d, fpgaver = %x.%x\n", wc->numspans, wc->fpgaver>>8, wc->fpgaver&0xff);
						break;
					default:
						printk("Hardware ID desconhecido! Abortando...\n");										
						kfree(wc);						
						return -EIO;
						break;
				}
				wc->variety = dt->desc;
			} else {
				printk("Card is NOT a Intelbras PXE card! (ID = 0x%08x)\n", *wc->membase);
				kfree(wc);
				return -EIO;
			}
			
			// Reseta a(s) LIU(s)
			acte4xx_liu_reset(wc);


			// testa interface SPI com a LIU
			/*if(acte4xx_spi_test(wc))
				printk("ERRO NO TESTE DA SPI!!!\n");
			else
				printk("Interface SPI OK!\n");*/
																
			/* This rids of the Double missed interrupt message after loading */
			wc->last0 = 1;

			printk("Found PXE%d%02d at base address %08lx, remapped to %p\n", wc->numspans, (wc->hwid&0x01)?00:45, wc->memaddr, wc->membase);

			wc->dev = pdev;

			wc->writechunk = 
				// 32 channels, Double-buffer, Read/Write, 4 spans
				(unsigned int *)pci_alloc_consistent(pdev, basesize * 2, &wc->writedma);
			if (!wc->writechunk) {
				printk("PXE%dxx: Unable to allocate DMA-able memory\n", wc->numspans);
				return -ENOMEM;
			}

			// Read is after the whole write piece (in words)
			wc->readchunk = wc->writechunk + basesize / 4;

			// Same thing but in bytes...  
			wc->readdma = wc->writedma + basesize;

			// Initialize Write/Buffers to all blank data 
			memset((void *)wc->writechunk,0x00, basesize);
			memset((void *)wc->readchunk,0xff, basesize);

			// Enable bus mastering
//			pci_set_master(pdev);

			/* Keep track of which device we are */
			pci_set_drvdata(pdev, wc);

			/* define se os canais sao T1 ou E1 (bitmap). Aqui todos forcados para E1
			   No futuro deve ser lido do hardware!!! */
			wc->t1e1 = 0x0f;
			
			/* inicializa contador de interrupcao */
			wc->intcount = 0;	
			
			for(x = 0; x < MAX_AC4_CARDS; x++) {
				if (!cards[x]) break;
			}

			if (x >= MAX_AC4_CARDS) {
				printk("No cards[] slot available!!\n");
				return -ENOMEM;
			}

			wc->num = x;
			cards[x] = wc;
			
			/* Allocate pieces we need here */
			for (x=0;x<4;x++) {
				if (wc->t1e1 & (1 << x)) {
					wc->tspans[x] = kmalloc(sizeof(struct acte4_span) + sizeof(struct zt_chan) * 31, GFP_KERNEL);
					if (wc->tspans[x]) {
						memset(wc->tspans[x], 0, sizeof(struct acte4_span) + sizeof(struct zt_chan) * 31);
						wc->tspans[x]->spantype = TYPE_E1;
					}
				} else {
					wc->tspans[x] = kmalloc(sizeof(struct acte4_span) + sizeof(struct zt_chan) * 24, GFP_KERNEL);
					if (wc->tspans[x]) {
						memset(wc->tspans[x], 0, sizeof(struct acte4_span) + sizeof(struct zt_chan) * 24);
						if (j1mode)
							wc->tspans[x]->spantype = TYPE_J1;
						else
							wc->tspans[x]->spantype = TYPE_T1;
					}
				}
				if (!wc->tspans[x])
					return -ENOMEM;
#ifdef ENABLE_WORKQUEUES
				INIT_WORK(&wc->tspans[x]->swork, workq_handlespan, wc->tspans[x]);
#endif				
				wc->tspans[x]->spanflags |= dt->flags;
			}

			if (request_irq(pdev->irq, acte4_interrupt, SA_INTERRUPT | SA_SHIRQ, "pxe", wc)) 
			{
				printk("PXE: Unable to request IRQ %d\n", pdev->irq);
				kfree(wc);
				return -EIO;
			}

			init_spans(wc);

			/* Launch cards as appropriate */
			x = 0;
			for(;;) {
				/* Find a card to activate */
				f = 0;
				for (x=0;cards[x];x++) {
					if (cards[x]->order <= highestorder) {
						acte4_launch(cards[x]);
						if (cards[x]->order == highestorder)
							f = 1;
					}
				}
				/* If we found at least one, increment the highest order and search again, otherwise stop */
				if (f) 
					highestorder++;
				else
					break;
			}

#ifdef CONFIG_PROC_FS
			sprintf(wc->proc_filename, "intelbras/PXE%d%02d_%d", wc->numspans, (wc->hwid&0x01)?00:45, wc->num+1);
			rw_proc_file = create_proc_entry(wc->proc_filename, 0666, NULL);
			if(rw_proc_file == NULL) {
				proc_mkdir("intelbras", NULL);
				rw_proc_file = create_proc_entry(wc->proc_filename, 0666, NULL);
				if(rw_proc_file == NULL)
					printk("Unable to create /proc/intelbras entry!!\n");
			} 
			if(rw_proc_file) {
				rw_proc_file->read_proc = acte4_read_proc;
				rw_proc_file->write_proc = acte4_write_proc;
				rw_proc_file->owner = THIS_MODULE;
				rw_proc_file->mode = S_IFREG | S_IRUGO;
				rw_proc_file->uid = 0;
				rw_proc_file->gid = 0;
				rw_proc_file->size = 0;			
				rw_proc_file->data = (void *) wc;
			}
#endif

			printk("Found a Wildcard: %s (SN: %d)\n", wc->variety, acte4xx_get_serial(pdev, wc));
			
#ifdef TIMER_DEBUG			
			// dispara timer de debug
			init_timer(&acte4xx_opt_timer);
			acte4xx_opt_timer.function = acte4xx_opt_timeout;
			acte4xx_opt_timer.data = (unsigned long) pdev;
			acte4xx_opt_timer.expires = jiffies + 100;
			add_timer(&acte4xx_opt_timer);
#endif
			res = 0;
		} else
			res = -ENOMEM;
	}
	return res;
}

static int acte4_hardware_stop(struct ac4 *wc)
{
	pd3_debugk("\n");
	/* Turn off DMA, leave interrupts enabled */
	wc->stopdma = 1;

	current->state = TASK_UNINTERRUPTIBLE;
	schedule_timeout((25 * HZ) / 1000);

	printk("\nStopped PXE%d%02d\n", wc->numspans, (wc->hwid&0x01)?00:45);
	return 0;
}



static void __devexit acte4_remove_one(struct pci_dev *pdev)
{
	struct ac4 *wc = pci_get_drvdata(pdev);
	int x;
	pd3_debugk("\n");
	if (wc) {

		pd3_debugk("desabilitando interrupcao!\n");
		*(wc->pci_vm_addr+0x13) &= 0xFFFF0000;		//	desabilita interrupcao	
			
		/* Stop hardware */
		acte4_hardware_stop(wc);

		/* Unregister spans */
		if (wc->tspans[0]->span.flags & ZT_FLAG_REGISTERED)
			zt_unregister(&wc->tspans[0]->span);
		if (wc->numspans > 1) {
			if (wc->tspans[1]->span.flags & ZT_FLAG_REGISTERED)
				zt_unregister(&wc->tspans[1]->span);
		}
		if (wc->numspans == 4) {
			if (wc->tspans[2]->span.flags & ZT_FLAG_REGISTERED)
				zt_unregister(&wc->tspans[2]->span);
			if (wc->tspans[3]->span.flags & ZT_FLAG_REGISTERED)
				zt_unregister(&wc->tspans[3]->span);
		}
#ifdef ENABLE_WORKQUEUES
		if (wc->workq) {
			flush_workqueue(wc->workq);
			destroy_workqueue(wc->workq);
		}
#endif			
#if 0
		/* Stop any DMA */
		__t1xxp_stop_dma(wc);

		/* In case hardware is still there */
		__t1xxp_disable_interrupts(wc);
		
		t1xxp_stop_stuff(wc);
#endif

#ifdef TIMER_DEBUG
		del_timer(&acte4xx_opt_timer);
#endif
		
		if (wc->membase)
			iounmap((void *)wc->membase);
		if(wc->pci_vm_addr)
			iounmap((void *)wc->pci_vm_addr);
			
		pci_release_regions(pdev);


#if 0
		if (wc->memaddr)
			release_mem_region(wc->memaddr, wc->memlen);
#endif

		/* Immediately free resources */
		pci_free_consistent(pdev, ZT_MAX_CHUNKSIZE * 2 * 2 * 32 * 4, (void *)wc->writechunk, wc->writedma);
#if 1
		free_irq(pdev->irq, wc);
#endif		
		cards[wc->num] = NULL;
		pci_set_drvdata(pdev, NULL);
		for (x=0;x<wc->numspans;x++) {
			if (wc->tspans[x])
				kfree(wc->tspans[x]);
		}
#ifdef CONFIG_PROC_FS
		remove_proc_entry(wc->proc_filename, NULL);
#endif
		kfree(wc);
	}
}


static struct pci_device_id acte4_pci_tbl[] __devinitdata =
{
	{ PCI_DEVICE(PCI_VENDOR_ID_PLX, PCI_DEVICE_ID_PLX_9030), },
	{ 0, }
};


static struct pci_driver acte4_driver = {
	name: 	"Unified pxe driver",
	probe: 	acte4_init_one,
#ifdef LINUX26
	remove:	__devexit_p(acte4_remove_one),
#else
	remove:	acte4_remove_one,
#endif
	suspend: NULL,
	resume:	NULL,
	id_table: acte4_pci_tbl,
};

static int __init acte4_init(void)
{
	int res;
	printk("Unified Intelbras PXE PCI Card Driver\n");	
	res = zap_pci_module(&acte4_driver);
	if (res) {
		return -ENODEV;
	}
	return 0;
}

static void __exit acte4_cleanup(void)
{
	printk("Unified Intelbras PXE PCI Card Driver Cleanup\n");
	pci_unregister_driver(&acte4_driver);
}


MODULE_AUTHOR("Intelbras (rvaliati@pd3.com.br)");
MODULE_DESCRIPTION("Unified Intelbras PXE PCI Card Driver");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
#ifdef LINUX26
module_param(debug, int, 0600);
module_param(loopback, int, 0600);
module_param(noburst, int, 0600);
module_param(debugslips, int, 0600);
module_param(polling, int, 0600);
module_param(timingcable, int, 0600);
module_param(t1e1override, int, 0600);
module_param(alarmdebounce, int, 0600);
module_param(j1mode, int, 0600);
#else
MODULE_PARM(debug, "i");
MODULE_PARM(loopback, "i");
MODULE_PARM(noburst, "i");
MODULE_PARM(debugslips, "i");
MODULE_PARM(polling, "i");
MODULE_PARM(timingcable, "i");
MODULE_PARM(t1e1override, "i");
MODULE_PARM(alarmdebounce, "i");
MODULE_PARM(j1mode, "i");
#endif

MODULE_DEVICE_TABLE(pci, acte4_pci_tbl);

module_init(acte4_init);
module_exit(acte4_cleanup);
